/* Tracer.cpp -- wait-free tracer for Firebird.
 *
 * The contents of this file are subject to the Interbase Public License
 * Version 1.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 * http://www.firebirdsql.org/index.php?op=doc&id=ipl
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 * 
 * The Original Code was created by Kerio Technologies.
 * 
 * Copyright 2008 Kerio Technologies Inc. All rights reserved.
 * 
 */
#include "Tracer.h"
#include "../common/classes/timestamp.h"

PageTracer pageTracer;

//-------------------------------------------------------------
// Helper.

namespace detail {

	std::string getModulePath()
	{
		std::string result(MAX_PATH, ' ');
		std::string::size_type slash;

		const DWORD filenameSize = GetModuleFileNameA(NULL, &(result[0]), static_cast<DWORD>(result.size()));
		result.resize(filenameSize);

		if ((slash = result.find_last_of("/\\")) != std::string::npos) {
			result.resize(slash);
		}

		return result;
	}

};

//-------------------------------------------------------------
// Init/destruct.

PageTracer::PageTracer() :	lastSequenceNumber(0), 
							numberOfFreeBufferItems(bufferItems), 
							tracingEnabled(false),
							wakeConsumerPort(NULL),
							logFileHandle(INVALID_HANDLE_VALUE)
{
	// Initialize whether we are tracing.
	this->tracingEnabled = this->shouldTrace();

	// Is tracing enabled? If not, no further initialization is necessary.
	if (!this->isTracing()) {
		return;
	}

	try {

		// Queue of empty items, initially full.
		pInitializeSListHead(&this->emptyListHead);

		for (size_t i = 0; i < bufferItems; ++i) {
			pInterlockedPushEntrySList(&this->emptyListHead, &( this->traceBuffer[i].slistEntry )); 
		}

		// Queue of full items, initially empty.
		pInitializeSListHead(&this->fullListHead);

		// IO port for occasional attempts to wake up the consumer.
		this->wakeConsumerPort = CreateIoCompletionPort(
			INVALID_HANDLE_VALUE,					
			NULL,
			NULL,
			1	// NumberOfConcurrentThreads
			);

		if (this->wakeConsumerPort == NULL) {
			throw std::runtime_error("CreateIoCompletionPort failed");
		}

		// Create a log file.
		std::string logFilePath = detail::getModulePath() + "\\firebird-pg.log";

		this->logFileHandle = CreateFileA(logFilePath.c_str(), GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
		if (this->logFileHandle == INVALID_HANDLE_VALUE) {
			throw std::runtime_error("CreateFileA failed");
		}
		
		// Move to end-of-file and log a separation line.
		SetFilePointer(this->logFileHandle, 0, NULL, FILE_END);
		
		const char* startOfTrace = "--\r\n";
		DWORD unusedBytesWritten;
		WriteFile(this->logFileHandle, startOfTrace, strlen(startOfTrace), &unusedBytesWritten, NULL);

		// Start the consumer thread.
		this->loggerThreadHandle = reinterpret_cast<HANDLE>(_beginthread(PageTracer::runConsumerThread, 0, this));
		if (this->loggerThreadHandle == reinterpret_cast<HANDLE>(-1L)) {
			throw std::runtime_error("_beginthread failed");
		}
	} 
	catch (std::exception&) {

		this->tracingEnabled = false;

		if (this->wakeConsumerPort != NULL) {
			CloseHandle(this->wakeConsumerPort);
		}

		if (this->logFileHandle != INVALID_HANDLE_VALUE) {
			CloseHandle(this->logFileHandle);
		}
	}
}

PageTracer::~PageTracer()
{
	// If tracing is not enabled, do nothing.
	if (!this->isTracing()) {
		return;
	}

	// Signal the consumer.
	TraceItem endTraceItem;
	endTraceItem.itemType = StopTracing;

	this->postTraceEvent(endTraceItem);

	// Wait until the consumer finishes.
	WaitForSingleObject(this->loggerThreadHandle, INFINITE);

	// Close the trace file.
	CloseHandle(this->logFileHandle);
}

//-------------------------------------------------------------
// Initialization.

bool PageTracer::getSListMethodAddresses()
{
	bool sListMethodsAvailable = false;

	do {

		HINSTANCE kernel32DllHandle = LoadLibraryA("kernel32.dll");
		if (kernel32DllHandle == NULL) {
			break;
		}

		pInitializeSListHead = (initializeSListHead_t)GetProcAddress(kernel32DllHandle, "InitializeSListHead");
		if (pInitializeSListHead == NULL) {
			break;
		}

		pInterlockedPushEntrySList = (interlockedPushEntrySList_t)GetProcAddress(kernel32DllHandle, "InterlockedPushEntrySList");
		if (pInterlockedPushEntrySList == NULL) {
			break;
		}

		pInterlockedPopEntrySList = (interlockedPopEntrySList_t)GetProcAddress(kernel32DllHandle, "InterlockedPopEntrySList");
		if (pInterlockedPopEntrySList == NULL) {
			break;
		}

		sListMethodsAvailable = true;
	} while (false);

	return sListMethodsAvailable;
}

bool PageTracer::shouldTrace()
{
	// The SList API is available since Windows XP.
	if (!getSListMethodAddresses()) {
		return false;
	}

	// Firebird source file "config/ConfigFile.cpp" also defines these constants locally. :-(
	const HKEY enableLogHKey = HKEY_LOCAL_MACHINE;
	const LPCWSTR enableLogSubkey = L"SOFTWARE\\Firebird Project\\Firebird Server\\Instances";
		
	// Setting PageTrace to 1 enables logging.
	const LPCWSTR enableLogValueName = L"PageTrace";
	HKEY openHKey;
	bool pageTraceLogEnabled = false;

	if (RegOpenKeyExW(enableLogHKey, enableLogSubkey, 0, KEY_READ, &openHKey) == ERROR_SUCCESS) {
		DWORD keyType;
		DWORD keyValue;
		DWORD keyValueLength = sizeof(keyValue);
			
		if (RegQueryValueExW(openHKey, enableLogValueName, 0, &keyType, reinterpret_cast<LPBYTE>(&keyValue),	&keyValueLength)
			== ERROR_SUCCESS && keyType == REG_DWORD) {
			pageTraceLogEnabled = (keyValue != 0);
		}
		RegCloseKey(openHKey);
	}

	return pageTraceLogEnabled;
}

//-------------------------------------------------------------
// Producer.

void PageTracer::postTraceEvent(const TraceItem& traceItem)
{
	// If tracing is not enabled, do nothing.
	if (!this->isTracing()) {
		return;
	}

	// Retrieve an empty slot.
	PSLIST_ENTRY slotBase;
	ULONG producerWaits = 0;
	const time_t now = time(NULL);

	while ((slotBase = pInterlockedPopEntrySList(&this->emptyListHead)) == NULL) {

		// Wait until there is a reasonable number of buffers.
		// This causes the context switches we are trying to avoid.
		while (numberOfFreeBufferItems < bufferLowWaterMark) {
			++producerWaits;

			PostQueuedCompletionStatus(
				this->wakeConsumerPort,
				0,
				NULL,
				NULL
				);

			Sleep(20);
		}
	}

	BufferItem_t& slot = *reinterpret_cast<BufferItem_t*>(slotBase);
	ULONG freeBufferItems = InterlockedDecrement(&this->numberOfFreeBufferItems);

	// Fill the slot.
	slot.sequenceNumber = InterlockedIncrement(&this->lastSequenceNumber);
	slot.producerWaits = producerWaits;
	slot.freeBufferItems = freeBufferItems;
	slot.time = now;
	slot.threadId = GetCurrentThreadId();
	slot.traceItem = traceItem;

	// Post the full slot.
	pInterlockedPushEntrySList(&this->fullListHead, &slot.slistEntry);

	// Stopping the consumer should be processed as fast as possible.
	if (traceItem.itemType == StopTracing) {
		PostQueuedCompletionStatus(
			this->wakeConsumerPort,
			0,
			NULL,
			NULL
			);
	}
}

//-------------------------------------------------------------
// Consumer.

void PageTracer::traceLoggerThread()
{
	bool finishFlag = false;
	char prefix[1024];
	char buffer[1024];

	while (true) {

		// Retrieve a full slot.
		PSLIST_ENTRY slotBase;

		while ((slotBase = pInterlockedPopEntrySList(&this->fullListHead)) == NULL) {

			// Finish if there are no more items to log.
			if (finishFlag) {
				return;
			}

			// Wait.
			DWORD unusedNumberOfBytes;
			ULONG unusedCompletionKey;
			LPOVERLAPPED unusedOverlapped;

			GetQueuedCompletionStatus(this->wakeConsumerPort, 
				&unusedNumberOfBytes, 
				&unusedCompletionKey, 
				&unusedOverlapped, 
				consumerTimeout
			);
		}

		BufferItem_t& slot = *reinterpret_cast<BufferItem_t*>(slotBase);

		// Process it.
		if (slot.traceItem.itemType == StopTracing) {
			finishFlag = true;
		}
		else {
            const int SECS_PER_HOUR	= 60 * 60;
            const int SECS_PER_DAY	= SECS_PER_HOUR * 24;

			// I had to add 1, otherwise it would print yesterday's date.
            const int days = (slot.time / SECS_PER_DAY) + 1;
            int rem = slot.time % SECS_PER_DAY;

            tm today;
            Firebird::TimeStamp::decode_date(days + Firebird::TimeStamp::GDS_EPOCH_START, &today);
            today.tm_hour = rem / SECS_PER_HOUR;
            rem %= SECS_PER_HOUR;
            today.tm_min = rem / 60;
            today.tm_sec = rem % 60;

			// Prepare the common prefix.
			sprintf(prefix, "%010u %04u-%02u-%02uT%02u:%02u:%02u %5u [%u %u]", 
                slot.sequenceNumber,
                today.tm_year + 1900, today.tm_mon, today.tm_mday, today.tm_hour, today.tm_min, today.tm_sec,
                slot.threadId,
                slot.producerWaits, 
                slot.freeBufferItems);

			// Message.
			switch (slot.traceItem.itemType) {
			case TraceNodeSplit:
				sprintf(buffer, "%s page %6ld splitted. split %6ld, right %6ld, parent %6ld\r\n",
					prefix,
					slot.traceItem.nodeSplit.pageNumber, 
					slot.traceItem.nodeSplit.split,
					slot.traceItem.nodeSplit.rightPage,
					slot.traceItem.nodeSplit.parentPage
					);
				break;
				
			case TraceNodeRemoved:
				sprintf(buffer, "%s node with page %6ld removed from parent page %6ld\r\n",
					prefix,
					slot.traceItem.nodeRemoved.pageNumber, 
					slot.traceItem.nodeRemoved.parentPage
					);
				break;

			case TracePageRemoved:
				sprintf(buffer, "%s page %6ld is removed from index. parent %6ld, left %6ld, right %6ld\r\n",
					prefix,
					slot.traceItem.pageRemoved.pageNumber,
					slot.traceItem.pageRemoved.parentPage,
					slot.traceItem.pageRemoved.leftPage,
					slot.traceItem.pageRemoved.rightPage
					);
				break;
				
			case TraceAllocPage:
            case TraceFreePage:
                {
                    std::string sourcePath(slot.traceItem.allocOrFreePage.srcFile);
                    std::string::size_type pos = sourcePath.find_last_of('\\');
                    if (pos == std::string::npos) {
                        pos = 0;
                    }

                    sprintf(buffer, "%s page %6ld %s at %s : %4d\r\n",
                        prefix,
                        slot.traceItem.allocOrFreePage.pageNumber,
                        (slot.traceItem.itemType == TraceAllocPage)? "allocated" : "released",
                        sourcePath.c_str() + pos + 1,
                        slot.traceItem.allocOrFreePage.srcLine
                        );
				}
				break;
				
			default:
				sprintf(buffer, "%s INVALID %u\r\n",
                    prefix,
                    slot.traceItem.itemType
                    );
                break;
			}

			// Write it.
			DWORD unusedBytesWritten;
			WriteFile(this->logFileHandle, buffer, static_cast<DWORD>(strlen(buffer)), &unusedBytesWritten, NULL);
		}

		// Post the empty slot.
		pInterlockedPushEntrySList(&this->emptyListHead, &slot.slistEntry);
		InterlockedIncrement(&this->numberOfFreeBufferItems);
	}
}
