/* Tracer.h -- wait-free tracer for Firebird. Heavily dependent on Windows API.
 *
 * The contents of this file are subject to the Interbase Public License
 * Version 1.0 (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 * http://www.firebirdsql.org/index.php?op=doc&id=ipl
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 * 
 * The Original Code was created by Kerio Technologies.
 * 
 * Copyright 2008 Kerio Technologies Inc. All rights reserved.
 * 
 */

#pragma once
#include <string>
#include <windows.h>
#include <time.h>
#include <process.h>
#include "../include/fb_types.h"

#if defined(SUPERSERVER) && defined(EMBEDDED)
#define DEBUG_BTR_SPLIT
#endif

class PageTracer {
public:
	enum TraceItemType {
		StopTracing,
		TraceNodeSplit,
		TraceNodeRemoved,
		TracePageRemoved,
		TraceAllocPage,
		TraceFreePage
	};

	typedef struct TraceItem {
		TraceItemType itemType;
		union {
			// "page %6ld splitted. split %6ld, right %6ld, parent %6ld"
			struct {
				SLONG pageNumber;
				SLONG split;
				SLONG rightPage;
				SLONG parentPage;
			} nodeSplit;

			// "node with page %6ld removed from parent page %6ld" (two places)
			struct {
				SLONG pageNumber;
				SLONG parentPage;
			} nodeRemoved;

			// "page %6ld is removed from index. parent %6ld, left %6ld, right %6ld"
			struct {
				SLONG pageNumber;
				SLONG parentPage;
				SLONG leftPage;
				SLONG rightPage;
			} pageRemoved;
			
			// "page %6ld allocated at %s : %4d"
			// "page %6ld released  at %s : %4d"
			struct {
                SLONG pageNumber;
                const char* srcFile;
                int srcLine;
			} allocOrFreePage;
		};
	} TraceItem_t;

	PageTracer();

	~PageTracer();

	bool isTracing()
	{
		return this->tracingEnabled;
	}

	void postTraceEvent(const TraceItem_t& traceItem);
	
	ULONG getEventSequence()
	{
        return InterlockedIncrement(&this->lastSequenceNumber);
	}

private:
	void traceLoggerThread();

	static void __cdecl runConsumerThread(void *arg)
	{
		PageTracer* const thisTracer = reinterpret_cast<PageTracer *>(arg);
		if (thisTracer != NULL) {
			thisTracer->traceLoggerThread();
		}
	}

private:
	bool getSListMethodAddresses();
	bool shouldTrace();

private:
	static const size_t bufferItems = 8 * 1024;					// 8k items for now
	static const size_t bufferLowWaterMark = bufferItems/2;		// when to wakeup the consumer
	static const DWORD consumerTimeout = 20 * 1000;				// consumer timeout is 20 seconds

	// Set once at FB start.
	bool tracingEnabled;

	// Heads of the Windows wait-free lists.
	SLIST_HEADER emptyListHead;
	SLIST_HEADER fullListHead;

	// Definitions and pointers to SList methods.
	typedef VOID (WINAPI* initializeSListHead_t)(PSLIST_HEADER);
	typedef PSLIST_ENTRY (WINAPI* interlockedPushEntrySList_t)(PSLIST_HEADER, PSLIST_ENTRY);
	typedef PSLIST_ENTRY (WINAPI* interlockedPopEntrySList_t)(PSLIST_HEADER);

	initializeSListHead_t pInitializeSListHead;
	interlockedPushEntrySList_t pInterlockedPushEntrySList;
	interlockedPopEntrySList_t pInterlockedPopEntrySList;

	typedef struct BufferItem {
		SLIST_ENTRY slistEntry;
		ULONG sequenceNumber;
		ULONG producerWaits;
		ULONG freeBufferItems;
		time_t time;
		DWORD threadId;
		TraceItem_t traceItem;
	} BufferItem_t;

	
	BufferItem_t traceBuffer[bufferItems];	// Buffer from which both lists are allocated.

	// Used by producer.
	volatile LONG lastSequenceNumber;
	volatile LONG numberOfFreeBufferItems;
	
	HANDLE wakeConsumerPort;				// Used to wake up the consumer on low space condition.
	HANDLE loggerThreadHandle;				// Consumer thread.
	HANDLE logFileHandle;					// Log file.
};

extern PageTracer pageTracer;
